/**************************************************************************************
 3D Visualization: Food & Drink Words                
 Tested in Processing 3                                    
 
 Author: Chantal Nguyen                        
 Supervisor: George Legrady                           
 
 Purpose: Visualizes words found in titles of Dewey 641 media at the Seattle Public
 Library, projected into 3D space using t-SNE. Spatial proximity is determined by
 analyzing the contextual similarity between words. Toggling options described below 
 allow the viewer to visualize the word co-occurence network, where connections between 
 words are determined by the number of times two words co-occur in the same title. 
 This represents a much more simplistic way of obtaining relationships between words 
 than word2vec + t-SNE.
 
 Usage: 
 1.  A mouse left-drag will rotate the camera around the subject.
 2.  A right drag will zoom in and out. 
 3.  A middle-drag (command-left-drag on mac) will pan. 
 4.  A double-click restores the camera to its original position. 
 5.  The shift key constrains rotation and panning to one axis or the other.
 6.  Pressing C toggles color modes: 
     [default] where color indicates frequency of the respective word, from green 
     (less frequent) to red (more frequent)
     [community] where points of the same color are within the same community in
     the word co-occurence network
 7.  Pressing 1 will make visible all edges between nodes in the word co-
     occurrence network. Pressing 2-9 or Q-P (corresponding to 10-19) will threshold 
     the edge weight such that only edges with weight greater than or equal to the 
     entered value will be shown.
 8.  Pressing 0 will hide the edges (default view).
 9.  Hovering over a node will show the corresponding label and the 20 closest related
     words.
 10. Search bar in lower right corner: type a word to highlight the corresponding node 
     and show 20 closest related words (if word does not exist in dataset, nothing will
     happen)
 **************************************************************************************/

import peasy.*;
import controlP5.*;
PeasyCam cam;
ControlP5 cp5;
PFont f;

float[] rotations = new float[3];

int rows, cols;

Table label_table;
Table coord_table;
Table freq_table;
Table assoc_table;
Table adjacency_table;
Table community_table;

float[][] coord_matrix;
float[] freq_matrix;
int[][] sparse_adjacency;
int[] communities;

float max_x_coord, max_y_coord, max_z_coord, min_x_coord, min_y_coord, min_z_coord;
float max_freq;
float max_edge_weight;
int min_weight = 1000;

final int minBoxSize = -140;
final int maxBoxSize = 140;

ArrayList<word> words = new ArrayList<word>();

// labelmap maps the words to their (arbitrarily set) order in the csv
HashMap<String, Integer> labelmap = new HashMap<String, Integer>();

PImage colorBar;
String[] colorPath = {"Colormap4.jpg"};

color cc1 = #F46C66;
color cc2 = #4C8493;
color cc3 = #FAC856;
color cc4 = #6E987B;
color cc5 = #BB4F24;
color cc6 = #3D89A7;
color cc7 = #CDA91D;
color cc8 = #841F11;
color cc9 = #C7904E;
color cc10 = #335144;
color cc11 = #C38C66;
color cc12 = #D31719;
color cc13 = #3C3B54;
color cc14 = #D28F33;
color cc15 = #133D30;
color cc16 = #CEB3D3;

color[] colors = {cc1, cc2, cc3, cc4, cc5, cc6, cc7, cc8, cc9, cc10, cc11, cc12, cc13, cc14, cc15, cc16};
color c;

word selectedWord = null;
boolean toggle;

void setup() {
  size(1300, 850, P3D);
  smooth();
  cam = new PeasyCam(this, 325);
  cam.setWheelScale(0.1);

  float cameraZ = (height/2.0) / tan(PI*60.0/360.0);
  perspective(PI/3.5, width/(1.1*height), 0.001, cameraZ*10.0); //set near clip to 0.001 to go inside data

  f = createFont("Helvetica", 24);



  // load t-SNE coordinates
  coord_table = loadTable("coords.csv"); 
  rows = coord_table.getRowCount();
  cols = coord_table.getColumnCount();

  coord_matrix = new float[rows][cols]; // initialize array of coordinates

  // import table data into array
  for (int i=0; i<coord_matrix.length; i++) {
    for (int j=0; j<coord_matrix[0].length; j++) {
      coord_matrix[i][j] = coord_table.getFloat(i, j);
    }
  }

  min_x_coord = 10000;
  min_y_coord = 10000;
  min_z_coord = 10000;
  for (int i=0; i<coord_matrix.length; i++) {
    if (coord_matrix[i][0] > max_x_coord) {
      max_x_coord = coord_matrix[i][0];
    }
    if (coord_matrix[i][1] > max_y_coord) {
      max_y_coord = coord_matrix[i][1];
    }
    if (coord_matrix[i][2] > max_z_coord) {
      max_z_coord = coord_matrix[i][2];
    }
    if (coord_matrix[i][0] < min_x_coord) {
      min_x_coord = coord_matrix[i][0];
    }
    if (coord_matrix[i][1] < min_y_coord) {
      min_y_coord = coord_matrix[i][1];
    }
    if (coord_matrix[i][2] < min_z_coord) {
      min_z_coord = coord_matrix[i][2];
    }
  }

  // load frequencies of words
  freq_table = loadTable("frequencies.csv");
  rows = freq_table.getRowCount();

  freq_matrix = new float[rows];

  for (int i=0; i<freq_matrix.length; i++) {
    freq_matrix[i] = freq_table.getFloat(i, 0);
    if (freq_matrix[i] > max_freq) {
      max_freq = freq_matrix[i];
    }
  }

  // load adjacency matrix
  adjacency_table = loadTable("adjacency.csv");
  rows = adjacency_table.getRowCount();
  cols = adjacency_table.getColumnCount();

  sparse_adjacency = new int[rows][cols]; // initialize sparse adjacency matrix (only edges are listed)

  // import table data into array
  for (int i=0; i<sparse_adjacency.length; i++) {
    sparse_adjacency[i][0] = adjacency_table.getInt(i, 0) - 1; // source node (subtract 1 because they are numbered from 1 in the csv)
    sparse_adjacency[i][1] = adjacency_table.getInt(i, 1) - 1; // target node
    sparse_adjacency[i][2] = adjacency_table.getInt(i, 2); // edge weight
    if (sparse_adjacency[i][2] > max_edge_weight) {
      max_edge_weight = sparse_adjacency[i][2];
    }
  }

  // load community mappings
  community_table = loadTable("communities.csv");
  rows = community_table.getRowCount();

  communities = new int[rows]; // initialize community mappings

  // import table data into array
  for (int i=0; i<communities.length; i++) {
    {      
      communities[i] = community_table.getInt(i, 0);
    }
  }

  // get list of words
  label_table = loadTable("labels.csv");
  assoc_table = loadTable("associations.csv", "header");

  for (int i=0; i<coord_matrix.length; i++) {
    ArrayList<String> tempList = new ArrayList<String>();
    for (int j=0; j<20; j++) {
      tempList.add(assoc_table.getString(i*20+j, 1));
    }
    words.add(new word(label_table.getString(i, 0), coord_matrix[i][0], coord_matrix[i][1], coord_matrix[i][2], freq_matrix[i], communities[i], tempList));
    labelmap.put(label_table.getString(i, 0), i); // map words back to original ordering
  }

  cp5 = new ControlP5(this);

  // add textfield
  // user enters a word, if word is in vocab, then corresponding point will be highlighted
  Textfield selectWord = cp5.addTextfield("selectWord")
    .setPosition(width-130, height-30)
    .setSize(50, 10)
    .setFocus(true)
    .setColor(color(0))
    .setColorBackground(color(255))
    .setColorActive(color(255))
    .setColorForeground(color(255))
    .setCaptionLabel("Search vocab")
    ;
  selectWord.getCaptionLabel().getStyle().marginTop  = -14; 
  selectWord.getCaptionLabel().getStyle().marginLeft = 55;
  cp5.setAutoDraw(false);
}

void draw() {
  background(150);
  cam.beginHUD();
  cp5.draw();
  cam.endHUD();
  colorBar = loadImage(colorPath[0]);
  for (int i=0; i<sparse_adjacency.length; i++) {
    if (sparse_adjacency[i][2] >= min_weight) {
      float xVal1 = map(words.get(sparse_adjacency[i][0]).x_coord, min_x_coord, max_x_coord, minBoxSize, maxBoxSize);
      float yVal1 = map(words.get(sparse_adjacency[i][0]).y_coord, min_y_coord, max_y_coord, minBoxSize, maxBoxSize);
      float zVal1 = map(words.get(sparse_adjacency[i][0]).z_coord, min_z_coord, max_z_coord, minBoxSize, maxBoxSize);
      float xVal2 = map(words.get(sparse_adjacency[i][1]).x_coord, min_x_coord, max_x_coord, minBoxSize, maxBoxSize);
      float yVal2 = map(words.get(sparse_adjacency[i][1]).y_coord, min_y_coord, max_y_coord, minBoxSize, maxBoxSize);
      float zVal2 = map(words.get(sparse_adjacency[i][1]).z_coord, min_z_coord, max_z_coord, minBoxSize, maxBoxSize);

      beginShape(LINES);
      if (toggle) {
        stroke(colors[words.get(sparse_adjacency[i][0]).community-1], 50);
      } else {
        stroke(colorBar.get((int)map(log(words.get(sparse_adjacency[i][0]).frequency), 0, log(max_freq), 2, colorBar.width-2), colorBar.height/2), 50);
      }
      strokeWeight(map(sparse_adjacency[i][2], 1, max_edge_weight, 0.5, 10));
      vertex(xVal1, yVal1, zVal1);
      if (toggle) {
        stroke(colors[words.get(sparse_adjacency[i][1]).community-1], 50);
      } else {
        stroke(colorBar.get((int)map(log(words.get(sparse_adjacency[i][1]).frequency), 0, log(max_freq), 2, colorBar.width-2), colorBar.height/2), 50);
      }      
      vertex(xVal2, yVal2, zVal2);
      endShape();
    }
  }

  for (int i=0; i<words.size(); i++) {
    float xVal = map(words.get(i).x_coord, min_x_coord, max_x_coord, minBoxSize, maxBoxSize);
    float yVal = map(words.get(i).y_coord, min_y_coord, max_y_coord, minBoxSize, maxBoxSize);
    float zVal = map(words.get(i).z_coord, min_z_coord, max_z_coord, minBoxSize, maxBoxSize);
    if (toggle) {
      c = colors[words.get(i).community-1];
    } else {
      c = colorBar.get((int)map(log(words.get(i).frequency), 0, log(max_freq), 2, colorBar.width-2), colorBar.height/2);
    }
    pushMatrix();
    translate(xVal, yVal, zVal);
    strokeWeight((int)map(log(words.get(i).frequency), 0, log(max_freq), 1, 20));
    stroke(c);
    point(0, 0, 0);

    PVector mouse_pos = new PVector(mouseX, mouseY);
    PVector screen_pos = new PVector(screenX(0, 0, 0), screenY(0, 0, 0));
    float mouseObjectDist = mouse_pos.dist(screen_pos);
    popMatrix();

    if (mouseObjectDist < 3) {
      pushMatrix();
      translate(xVal, yVal, zVal);
      strokeWeight(1.5*(int)map(log(words.get(i).frequency), 0, log(max_freq), 1, 20)); // enlarge selected point
      point(0, 0, 0);
      popMatrix();
      selectedWord = words.get(i);
    }
  }
  showHighlightedWord(selectedWord);

  cam.beginHUD();
  textAlign(LEFT, BOTTOM);
  textFont(f, 14);
  fill(255);
  text("Press C to toggle colors, 1 to display network edges, 2-9 and Q-P to threshold edge weights, 0 to remove edges", 10, height-20);
  cam.endHUD();
}

class word {
  String label;
  float x_coord;
  float y_coord;
  float z_coord;
  float frequency;
  int community;
  ArrayList<String> associations;

  word(String label, float x_coord, float y_coord, float z_coord, float frequency, int community, ArrayList associations) {
    this.label = label;
    this.x_coord = x_coord;
    this.y_coord = y_coord;
    this.z_coord = z_coord;
    this.frequency = frequency;
    this.community = community;
    this.associations = associations;
  }
}

void showHighlightedWord(word selectedWord) {
  if (selectedWord != null) {
    float xV = map(selectedWord.x_coord, min_x_coord, max_x_coord, minBoxSize, maxBoxSize);
    float yV = map(selectedWord.y_coord, min_y_coord, max_y_coord, minBoxSize, maxBoxSize);
    float zV = map(selectedWord.z_coord, min_z_coord, max_z_coord, minBoxSize, maxBoxSize);  
    pushMatrix();
    translate(xV, yV, zV);
    strokeWeight(20);
    fill(255);
    textMode(SHAPE);
    textFont(f, 5);
    pushMatrix();
    rotations = cam.getRotations();
    rotateX(rotations[0]);
    rotateY(rotations[1]);
    rotateZ(rotations[2]);
    text(selectedWord.label, 2, 2);
    popMatrix();

    // print labels of 20 similar words at their respective points
    translate(-1*xV, -1*yV, -1*zV);
    String associationString = "";
    for (int j=0; j<selectedWord.associations.size(); j++) {
      pushMatrix();
      float xx = map(words.get(labelmap.get(selectedWord.associations.get(j))).x_coord, min_x_coord, max_x_coord, minBoxSize, maxBoxSize);
      float yy = map(words.get(labelmap.get(selectedWord.associations.get(j))).y_coord, min_y_coord, max_y_coord, minBoxSize, maxBoxSize);
      float zz = map(words.get(labelmap.get(selectedWord.associations.get(j))).z_coord, min_z_coord, max_z_coord, minBoxSize, maxBoxSize); 
      translate(xx, yy, zz);
      rotateX(rotations[0]);
      rotateY(rotations[1]);
      rotateZ(rotations[2]);
      textFont(f, 3);
      fill(242, 186, 184);
      text(words.get(labelmap.get(selectedWord.associations.get(j))).label, 2, 2);
      popMatrix();
      associationString = associationString + " " + words.get(labelmap.get(selectedWord.associations.get(j))).label;
    }

    // in top left corner, print highlighted label + 20 associations
    cam.beginHUD();
    textAlign(LEFT, TOP);
    textFont(f, 40);
    fill(255);
    text(selectedWord.label, 10, 10);
    textFont(f, 20);
    fill(242, 186, 184);
    text(associationString, 7, 50);
    cam.endHUD();
    popMatrix();
  }
}

public void selectWord(String theText) {
  // automatically receives results from controller input
  println("user has entered : "+theText);
  if (labelmap.containsKey(theText)) {
    selectedWord = words.get(labelmap.get(theText));
  }
}

void keyPressed() {
  if (key == 'C') {
    toggle = !toggle;
  } else if (key == '1') {
    min_weight = 1;
  } else if (key == '2') {
    min_weight = 2;
  } else if (key == '3') {
    min_weight = 3;
  } else if (key == '4') {
    min_weight = 4;
  } else if (key == '5') {
    min_weight = 5;
  } else if (key == '6') {
    min_weight = 6;
  } else if (key == '7') {
    min_weight = 7;
  } else if (key == '8') {
    min_weight = 8;
  } else if (key == '9') {
    min_weight = 9;
  } else if (key == '0') {
    min_weight = 1000;
  } else if (key == 'Q') {
    min_weight = 10;
  } else if (key == 'W') {
    min_weight = 11;
  } else if (key == 'E') {
    min_weight = 12;
  } else if (key == 'R') {
    min_weight = 13;
  } else if (key == 'T') {
    min_weight = 14;
  } else if (key == 'Y') {
    min_weight = 15;
  } else if (key == 'U') {
    min_weight = 16;
  } else if (key == 'I') {
    min_weight = 17;
  } else if (key == 'O') {
    min_weight = 18;
  } else if (key == 'P') {
    min_weight = 19;
  }
}
